/******************************************************************************
*
* This file contains code to emit tracing for debugging and profiling.
*
* Up to 32 different categories of debug messages may be defined, and by
* editing this file it is possible to select which of the categories are
* enabled. Obviously in production code all categories should be disabled.
*
* To enable tracing for messages of one or more categories:
* . Change TRACE_OFF to TRACE_ON for the categories of interest.
* Note that there are additional configuration options as documented below.
*
* To define a new category:
* . Chose a new name for the category
* . Chose an existing spare category (TRACE_SPAREn) and rename the two
*   occurrences of this constant below.
*
* To add a new trace message:
* . Place code of the form
*
*   dbgtrace (domain, (printf-style-params));
*
* Note the parens around the printf-style-params.
* Domain can be one or more of the categories | (or'd) together, or
* TRACE_ALL. The message will be emitted if any of these categories is
* enabled.
*
* For really verbose messages, instead use dbgvtrace (). This is only enabled
* if TRACE_VERBOSE is also set.
*
* Note that because mk_make does not currently determine dependencies 
* fully, if you edit this file you will need to mk_make clean before 
* rebuilding.
*
******************************************************************************/

#ifndef _DBGTRACE_H_
#define _DBGTRACE_H_

#ifndef WRITLOSS
#define WRITLOSS 0
#endif

typedef struct _CALLER{
	const char *file;
	uns line;
	const char *caller;
	const char *callee;
	uns count;
	struct _CALLER *left;
	struct _CALLER *right;
} CALLER;

#ifdef MSC
struct timeval;
int gettimeofday(struct timeval *tv, void *tz);
#endif // MSC

/*
 * Define the filename which will be created here.
 * %p, if specified, will be replaced by the process id.
 */
// #define TRACE_FILENAME	"apltrace%p.txt"
#define TRACE_FILENAME	"apltrace.txt"

/*
 * Set to non-zero to append to any existing contents in the trace file,
 * otherwise the trace file will be recreated when the interpreter is run.
 * This is somewhat meaningless if the filename contains %p.
 */
#define TRACE_APPEND	0

/*
 * Set to non-zero if the interpreter crashes. The trace file will be closed
 * after every write, but this will have a significant performance impact.
 */
#define TRACE_CRASH	1

/*
 * Set to non-zero to have each message prepended with the elapsed time since
 * the first message.
 */
#define TRACE_ELAPSED	0

/*
 * Set to non-zero to have each message prepended with the elapsed time since
 * the previous message.
 */
#define TRACE_INTERVAL	0

/*
 * Identifier message to place at the start of the trace file (if defined)
 */
// #define TRACE_ID "Your message here"

/*
 * Set to no-zero to enable really verbose messages 
 */
#define TRACE_VERBOSE	0

/*
 * Enable or disable categories of messages by setting TRACE_ON or TRACE_OFF
 * as appropriate
 */
#define TRACE_QFILE	(TRACE_OFF << 0)
#define TRACE_WRITLOSS	(WRITLOSS  << 1)
#define TRACE_QXML	(TRACE_OFF << 2)
#define TRACE_CALLS	(TRACE_OFF << 3)
#define TRACE_QRX	(TRACE_OFF << 4)
#define TRACE_DMX	(TRACE_OFF << 5)
#define TRACE_SPARE6	(TRACE_OFF << 6)
#define TRACE_SPARE7	(TRACE_OFF << 7)
#define TRACE_SPARE8	(TRACE_OFF << 8)
#define TRACE_SPARE9	(TRACE_OFF << 9)
#define TRACE_SPARE10	(TRACE_OFF << 10)
#define TRACE_SPARE11	(TRACE_OFF << 11)
#define TRACE_SPARE12	(TRACE_OFF << 12)
#define TRACE_SPARE13	(TRACE_OFF << 13)
#define TRACE_SPARE14	(TRACE_OFF << 14)
#define TRACE_SPARE15	(TRACE_OFF << 15)
#define TRACE_SPARE16	(TRACE_OFF << 16)
#define TRACE_SPARE17	(TRACE_OFF << 17)
#define TRACE_SPARE18	(TRACE_OFF << 18)
#define TRACE_SPARE19	(TRACE_OFF << 19)
#define TRACE_SPARE20	(TRACE_OFF << 20)
#define TRACE_SPARE21	(TRACE_OFF << 21)
#define TRACE_SPARE22	(TRACE_OFF << 22)
#define TRACE_SPARE23	(TRACE_OFF << 23)
#define TRACE_SPARE24	(TRACE_OFF << 24)
#define TRACE_SPARE25	(TRACE_OFF << 25)
#define TRACE_SPARE26	(TRACE_OFF << 26)
#define TRACE_SPARE27	(TRACE_OFF << 27)
#define TRACE_SPARE28	(TRACE_OFF << 28)
#define TRACE_SPARE29	(TRACE_OFF << 29)
#define TRACE_SPARE30	(TRACE_OFF << 30)
#define TRACE_SPARE31	(TRACE_OFF << 31)

#define WANT_TRACING ( 0	\
	| TRACE_QFILE		\
	| TRACE_WRITLOSS	\
	| TRACE_QXML		\
	| TRACE_CALLS		\
	| TRACE_QRX		\
	| TRACE_DMX		\
	| TRACE_SPARE6		\
	| TRACE_SPARE7		\
	| TRACE_SPARE8		\
	| TRACE_SPARE9		\
	| TRACE_SPARE10		\
	| TRACE_SPARE11		\
	| TRACE_SPARE12		\
	| TRACE_SPARE13		\
	| TRACE_SPARE14		\
	| TRACE_SPARE15		\
	| TRACE_SPARE16		\
	| TRACE_SPARE17		\
	| TRACE_SPARE18		\
	| TRACE_SPARE19		\
	| TRACE_SPARE20		\
	| TRACE_SPARE21		\
	| TRACE_SPARE22		\
	| TRACE_SPARE23		\
	| TRACE_SPARE24		\
	| TRACE_SPARE25		\
	| TRACE_SPARE26		\
	| TRACE_SPARE27		\
	| TRACE_SPARE28		\
	| TRACE_SPARE29		\
	| TRACE_SPARE30		\
	| TRACE_SPARE31		\
)

#define TRACE_ON	1
#define TRACE_OFF	0

#define TRACE_ALL	0xFFFFFFFF
	
#if (WANT_TRACING != 0)
extern void _dbgtrace (const char *fmt, ...);
extern void _dbgdumphex (const unsigned char *desc, const void *buf, unsigned int buflen);
extern void _dbgtrace_log_caller (const char *callee, const char *file, unsigned line, const char *function);
extern void _dbgtrace_term (void);

# define dbgtrace(V, x) do { if ((V) & WANT_TRACING) _dbgtrace x; } while (0)
# define dbgdumphex(V, D, B, L) do { if ((V) & WANT_TRACING) _dbgdumphex (D, B, L); } while (0)
# if TRACE_VERBOSE
#  define dbgvtrace(V, x) dbgtrace(V, x)
#  define dbgvdumphex(V, D, B, L) dbgdumphex(V, D, B, L)
# else
#  define dbgvtrace(V, x)
#  define dbgvdumphex(V, D, B, L)
# endif
# define dbgtrace_term() _dbgtrace_term ()
#else
# define dbgtrace(V, x)
# define dbgvtrace(V, x)
# define dbgdumphex(V, D, B, L)
# define dbgvdumphex(V, D, B, L)
# define dbgtrace_term()
#endif

#ifdef MSC
# define FMT32 "%I32i"
# define FMT64 "%I64i"
#else
# define FMT32 "%u"
# define FMT64 "%llu"
#endif
#if 64 == xxBIT
#define FMTBIT	FMT64
#else
#define FMTBIT	FMT32
#endif

#if (TRACE_CALLS)

# define TRACED_FN(N) _tf_##N
# define TRACED_FN_PARAMS \
	const char *__file, unsigned __line, const char *__function,
# define TRACE_CALLER() \
	_dbgtrace_log_caller(__FUNCTION__, __file, __line, __function)

# define TRACED_CALL_EXTRA __FILE__, __LINE__, __FUNCTION__

/*
 * To add tracing of calls to a particular function:
 * 1. Add a macro here to modify all calls to the function. In general, for
 *    function "fn" with params "p1, p2, p3" this would be
 *    #define fn(p1, p2, p3) TRACED_FN(fn) (TRACED_CALL_EXTRA, p1, p2, p3)
 *    and for getspace this would be
 *    #define getspace(nwds) TRACED_FN(getspace) (TRACED_CALL_EXTRA, nwds)
 * 2. Amend the function declaration and definition. For function "fn" with
 *    arguments "(t1 p1, t2 p2, t3 p3)" returning "t" this would be modified
 *    follows:
 *      declaration: t fn(t1 p1, t2 p2, t3 p3);
 *              -->  t TRACED_FN(fn)(TRACED_FN_PARAMS t1 p1, t2 p2, t3 p3);
 *      definition:  t fn(t1 p1, t2 p2, t3 p3) { ...
 *              -->  t TRACED_FN(fn)(TRACED_FN_PARAMS t1 p1, t2 p2, t3 p3) { ...
 *    so for getspace the these become
 *      extern POCKET *TRACED_FN(getspace)(TRACED_FN_PARAMS PLENGTH nwds);
 *      POCKET *TRACED_FN(getspace)(TRACED_FN_PARAMS PLENGTH nwds)
 * 3. Add the following line as the first statement in the function:
 *    TRACE_CALLER ();
 * 4. Enable tracing for the modified functions by setting TRACE_CALLS to
 *    TRACE_ON.
 */

# define getspace(nwds)	TRACED_FN(getspace) (TRACED_CALL_EXTRA, nwds)

#else // TRACE_CALLS

# define TRACED_FN(N) N
# define TRACED_FN_PARAMS
# define TRACE_CALLER()

#endif // TRACE_CALLS

// eg: dbgdmp (TRACE_QRX, "%d", Rarg->RANK);
#define dbgdump(V, F, N) dbgtrace (V, (#N ": " F, N))

#endif /* _DBGTRACE_H_ */
